"""
***
June 1 2023
The files `bamSitesToWig.py`, `cutsToWig.pl`, `smoothWig.pl` 
in the `rocco` directory are copied from the PEPATAC repository.

Permalink:
https://github.com/databio/pepatac/tree/9ee0b6c1251b1addae8265c12f16cbfeae76d489/tools

See below for copyright notice.
***

Copyright 2021 Nathan Sheffield

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
#!/usr/bin/env python

__author__ = ["Nathan C. Sheffield", "Jason Smith"]
__credits__ = []
__license__ = "BSD2"
__version__ = "0.3.1"
__email__ = "nathan@code.databio.org"

from argparse import ArgumentParser
import itertools # Used for nested region looping across reads
import numpy
from operator import methodcaller
import os
import subprocess
import sys
import logmuse
import pararead
import pysam
import pkg_resources

from pararead import  ParaReadProcessor

MODES = ["dnase", "atac"]

# A function object like this will be pickled by the parallel call to map,
# So it cannot contain huge files or the pickling will limit everything.
# For this reason I must rely on global vars for the big stuff.
class CutTracer(pararead.ParaReadProcessor):
    """
    A function object that holds parameters and can be called
    on different chromosomes. This extends the ParaReadProcessor object so
    that it can be run in parallel.
    """
    def __init__(self, reads_filename, chrom_sizes_file, temp_parent, nProc,
        limit, verbosity, shift_factor={"+":0, "-":-0}, variable_step=False,
        exactbw=False, scale=1, bedout=False, smoothbw=False,
        smooth_length=25, step_size=5, retain_temp=False, tail_edge=False):
        # The resultAcronym should be set for each class
        self.resultAcronym="cuttrace"
        self.chrom_sizes_file = chrom_sizes_file
        self.shift_factor = shift_factor

        # Saving a smooth bigwig doubles the processor use for each chrom, so we
        # need to run half as many chroms at a time
        if smoothbw and exactbw:
            _LOGGER.info("Cutting parallel chroms in half to accommodate two tracks.")
            nProc = max(int(nProc / 2), 1)

        super(CutTracer,self).__init__(path_reads_file=reads_filename, cores=nProc,
            action=self.resultAcronym, temp_folder_parent_path=temp_parent, 
            limit=limit, allow_unaligned=False,
            retain_temp=retain_temp)
        self.exactbw = exactbw
        self.scale = scale
        self.verbosity=verbosity
        if variable_step:
            self.variable_step = 1 # True
        else:
            self.variable_step = 0 # False
        self.bedout = bedout
        self.smoothbw = smoothbw
        self.smooth_length = smooth_length
        self.step_size = step_size
        self.tail_edge = tail_edge

        # Confirm that all the commands we will need are callable

        try:
            self.check_command("wigToBigWig")
            self.check_command("perl")
            self.check_command("bigWigCat")
        except AttributeError:
            # backwards compatibility with earlier versions of pararead that did
            # not have a check_command function
            pass

    def register_files(self):
        super(CutTracer, self).register_files()

    def unbuffered_write(self, txt):
        """ 
        Writes unbuffered output by flushing after each stdout.write call
        """
        sys.stdout.write(txt)
        sys.stdout.flush()

    def __call__(self, chrom):
        """
        Workhorse function of the method.
        This function takes a chrom, and processes the reads in that chromosome
        in the input bamfile
        @param chrom: a string with a chromosome, used by bamFile.fetch to
        grab a subset of reads from the bamfile
        """

        _LOGGER.debug(f"chrom: {chrom}")
        chrom_size = self.get_chrom_size(chrom)

        _LOGGER.debug("[Name: " + chrom + "; Size: " + str(chrom_size) + "]")
        reads = self.fetch_chunk(chrom)

        chromOutFile = self._tempf(chrom.replace("|","_"))
        chromOutFileBw = chromOutFile + ".bw"
        chromOutFileBwSm = chromOutFile + "_smooth.bw"
        tmpFile = chromOutFile + "_cuts.txt"

        cutsToWig = pkg_resources.resource_filename(__name__, 'cutsToWig.pl')
        cmd1 = ("sort -n | perl " + cutsToWig + " " + str(chrom_size) +
                " " + str(self.variable_step) + " " + str(self.scale))
        cmd2 = ("wigToBigWig -clip -fixedSummaries -keepAllChromosomes stdin " +
                self.chrom_sizes_file + " " + chromOutFileBw)
        _LOGGER.debug("  cutsToWigProcess: " + cmd1)
        _LOGGER.debug("  wigToBigWigProcess: " + cmd2)

        if self.exactbw:
            cutsToWigProcess = subprocess.Popen(cmd1, shell=True,
                stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            wigToBigWigProcess = subprocess.Popen(
                ['wigToBigWig', '-clip', '-fixedSummaries',
                 '-keepAllChromosomes', 'stdin',
                 self.chrom_sizes_file, chromOutFileBw],
                 stdin=cutsToWigProcess.stdout)

        if self.smoothbw:
            #cutsToWigSm = os.path.join(os.path.dirname(__file__),
                                       #"smoothWig.pl")
            cutsToWigSm =pkg_resources.resource_filename(__name__, 'smoothWig.pl')

            cmd1 = ("sort -n | tee " + tmpFile + " | perl " + cutsToWigSm +
                    " " + str(chrom_size) + " " +  str(self.smooth_length) +
                    " " + str(self.step_size) + " " + str(self.variable_step) +
                    " " + str(self.scale))
            cmd2 = ("wigToBigWig -clip -fixedSummaries " +
                    "-keepAllChromosomes stdin " + self.chrom_sizes_file +
                    " " + chromOutFileBwSm)
            cutsToWigProcessSm = subprocess.Popen(cmd1, shell=True,
                stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            wigToBigWigProcessSm = subprocess.Popen(
                ['wigToBigWig', '-clip', '-fixedSummaries',
                 '-keepAllChromosomes', 'stdin',
                 self.chrom_sizes_file, chromOutFileBwSm],
                 stdin=cutsToWigProcessSm.stdout)

        if self.bedout:
            chromOutFileBed = chromOutFile + ".bed"
            bedOut = open(chromOutFileBed, "w")


        def get_shifted_pos(read, shift_factor):
            """
            Shifts a read according to a shift factor to account for either
            transposition insertion site shifting or DNAse read shifting,
            depending on the strand of the read. Returns the shifted position of
            interest.
            :param read: A pysam read object
            :param shift_factor: A dict with positive or negative integer values
                for keys ["+", "-"], indicating how much (and which direction)
                to shift reads on the + or - strand
            """
            # default
            shifted_pos = None
            if not self.tail_edge:
                if read.flag & 1:  # paired
                    if read.flag == 99:  # paired, mapped in pair, mate reversed, first in pair
                        shifted_pos = read.reference_start + shift_factor["+"]
                        #r.reference_length  # col 8
                    elif read.flag == 147:  # mate of 99
                        shifted_pos = read.reference_end + shift_factor["-"]
                    elif read.flag == 163:  # paired, mapped in pair, mate reversed, second in pair
                        shifted_pos = read.reference_start + shift_factor["+"]
                    elif read.flag == 83:   # mate of 163
                        shifted_pos = read.reference_end + shift_factor["-"]
                else:  # unpaired
                    if read.flag & 16:  # read reverse strand
                        shifted_pos = read.reference_end + shift_factor["-"]
                    else:
                        shifted_pos = read.reference_start + shift_factor["+"]
            else: # Take 3' end of read
                if read.flag & 1:  # paired
                    if read.flag == 99:  # paired, mapped in pair, mate reversed, first in pair
                        shifted_pos = read.reference_end + shift_factor["-"]
                        #r.reference_length  # col 8
                    elif read.flag == 147:  # mate of 99
                        shifted_pos = read.reference_start + shift_factor["+"]
                    elif read.flag == 163:  # paired, mapped in pair, mate reversed, second in pair
                        shifted_pos = read.reference_end + shift_factor["-"]
                    elif read.flag == 83:   # mate of 163
                        shifted_pos = read.reference_start + shift_factor["+"]
                else:  # unpaired
                    if read.flag & 16:  # read reverse strand
                        shifted_pos = read.reference_start + shift_factor["+"]
                    else:
                        shifted_pos = read.reference_end + shift_factor["-"]

            return shifted_pos


        begin = 1

        if self.exactbw:
            if self.variable_step:
                header_line = ("variableStep chrom=" + chrom + "\n")
            else: 
                header_line = ("fixedStep chrom=" + chrom + " start=" +
                               str(begin) + " step=1\n")

            cutsToWigProcess.stdin.write(header_line.encode('utf-8'))

        if self.smoothbw:
            if self.variable_step:
                header_line = ("variableStep chrom=" + chrom + "\n")
            else:
                header_line = ("fixedStep chrom=" + chrom + " start=" +
                               str(begin) + " step=" + str(self.step_size) +
                               "\n")

            cutsToWigProcessSm.stdin.write(header_line.encode('utf-8'))

        try:
            for read in reads:
                shifted_pos = get_shifted_pos(read, shift_factor)

                if self.exactbw and shifted_pos:
                    cutsToWigProcess.stdin.write((str(shifted_pos) + "\n").encode('utf-8'))

                if self.smoothbw and shifted_pos:
                    cutsToWigProcessSm.stdin.write((str(shifted_pos) + "\n").encode('utf-8'))

                if self.bedout and shifted_pos:
                    strand = "-" if read.is_reverse else "+"
                    # The bed file needs 6 columns (even though some are dummy) 
                    # because MACS says so.
                    bedOut.write("\t".join([
                        chrom,
                        str(shifted_pos - self.smooth_length),
                        str(shifted_pos + self.smooth_length), 
                        "N", 
                        "0",
                        strand]) + "\n")
            # For chroms with no reads, the 'read' variable will not be bound.
            #if read in locals():

            # Clean up processes
            if self.exactbw:
                cutsToWigProcess.stdin.close()
                _LOGGER.debug("Encoding exact bigwig for " + chrom + 
                              " (last read position:" + str(read.pos) + ")...")
                wigToBigWigProcess.communicate()

            if self.bedout:
                bedOut.close()

            if self.smoothbw:
                cutsToWigProcessSm.stdin.close()
                _LOGGER.debug("Encoding smooth bigwig for " + chrom +
                              " (last read position:" + str(read.pos) + ")...")
                wigToBigWigProcessSm.communicate()

        except StopIteration as e:
            print("StopIteration error for chrom ", chrom, ": ", e)
            raise e

        return chrom


    def combine(self, good_chromosomes):
        """
        After running the process in parallel, this 'reduce' step will simply
        merge all the temporary files into one, and rename it to the output
        file name.
        """
        if not good_chromosomes:
            _LOGGER.info("No successful chromosomes, so no combining.")
            return
        elif len(good_chromosomes) == 1:
            subprocess.call(["mv",
                            self._tempf(good_chromosomes[0].replace("|","_")) +
                             ".bw", self.exactbw])
            subprocess.call(["mv",
                            self._tempf(good_chromosomes[0].replace("|","_")) +
                             "_smooth.bw", self.smoothbw])
        else:
            if self.exactbw:
                _LOGGER.info("Merging {} files into output file: '{}'".
                      format(len(good_chromosomes), self.exactbw))
                temp_files = [self._tempf(chrom) + ".bw" for chrom in good_chromosomes]
                files_exist = []
                for file in temp_files:
                    if os.path.isfile(file) and os.stat(file).st_size > 0:
                        files_exist.append(file)
                cmd = "bigWigCat " + self.exactbw + " " + " ".join(files_exist)
                _LOGGER.debug(cmd)
                p = subprocess.call(['bigWigCat', self.exactbw] + files_exist)

            if self.smoothbw:
                _LOGGER.info("Merging {} files into output file: '{}'".
                      format(len(good_chromosomes), self.smoothbw))
                temp_files = [self._tempf(chrom) + "_smooth.bw" for chrom in good_chromosomes]
                files_exist = []
                for file in temp_files:
                    if os.path.isfile(file) and os.stat(file).st_size > 0:
                        files_exist.append(file)
                cmd = "bigWigCat " + self.smoothbw + " " + " ".join(files_exist)
                _LOGGER.debug(cmd)
                p = subprocess.call(['bigWigCat', self.smoothbw] + files_exist)

            if self.bedout:
                # root, ext = os.path.splitext(self.exactbw)
                temp_files = [self._tempf(chrom) + ".bed" for chrom in good_chromosomes]
                files_exist = []
                for file in temp_files:
                    if os.path.isfile(file) and os.stat(file).st_size > 0:
                        files_exist.append(file)
                cmd = "cat " + " ".join(files_exist) + " > " + self.bedout
                _LOGGER.debug(cmd)
                p = subprocess.call(cmd, shell=True)


def parse_args(cmdl):
    parser = ArgumentParser(description='Bam processor')
    parser.add_argument('-i', '--infile', dest='infile',
        help="Input file (in bam or sam format)", required=True)
    parser.add_argument('-c', '--chrom-sizes-file',
        help="Chromosome sizes file", required=True)
    parser.add_argument('-v', '--variable-step', default=False, action='store_true',
        help="Use variableStep wiggle format. Default: fixedStep")
    parser.add_argument('-s', '--scale', dest='scale', default=1,
        help="Scale read count by this value. Default: 1")    
    parser.add_argument('-o', '--exactbw', dest='exactbw', default=None,
        help="Output filename for exact bigwig. Default: None")
    parser.add_argument('-w', '--smoothbw', dest='smoothbw', default=None,
        help="Output filename for smooth bigwig. Default: None")
    parser.add_argument('-r', '--step-size', default=5,
        help="Step size for smooth tracks. Default: 5")
    parser.add_argument('-b', '--bedout', default=None,
        help="Output filename for bed file. Default: None")
    parser.add_argument('-l', '--smooth-length',
        help="Smooth length for bed file", default=25, type=int)
    parser.add_argument('-d', '--tail-edge', action='store_true', default=False,
        help="Output the 3' end of the sequence read. Default: False")
    parser.add_argument('-m', '--mode', dest='mode', default=None, choices=MODES,
        help="Turn on DNase or ATAC mode (this adjusts the shift parameters)")
    parser.add_argument('-t', '--limit', dest='limit',
        help="Limit to these chromosomes", nargs = "+", default=None)
    parser.add_argument('-p', '--cores', dest='cores',
        help="Number of cores to use", default=2, type=int)
    parser.add_argument('-e', '--temp-parent',
        default="",#os.getcwd(),
        help="Temporary file location. By default it will use the working"
        " directory, but you can place this elsewhere if you'd like."
        " The actual folder will be based on the exactbw filename.")
    parser.add_argument('--retain-temp', action='store_true', default=False,
        help="Retain temporary files? Default: False")

    parser = logmuse.add_logging_options(parser)
    args = parser.parse_args(cmdl)
    if not (args.exactbw or args.smoothbw):
        parser.error('No output requested, use --exactbw and/or --smoothbw')
    return args

if __name__ == "__main__":

    args = parse_args(sys.argv[1:])
    _LOGGER = logmuse.logger_via_cli(args, make_root=True)

    if args.mode == "dnase":
        shift_factor = {"+":1, "-":0}  # DNase
    elif args.mode == "atac":
        shift_factor = {"+":4, "-":-5}  # ATAC
    else:
        shift_factor = {"+":0, "-":0}
    ct = CutTracer( reads_filename=args.infile,
                    chrom_sizes_file=args.chrom_sizes_file,
                    scale=args.scale,
                    variable_step=args.variable_step,
                    exactbw=args.exactbw,
                    smoothbw=args.smoothbw,
                    step_size=args.step_size,
                    bedout=args.bedout,
                    shift_factor=shift_factor,
                    smooth_length=args.smooth_length,
                    tail_edge=args.tail_edge,
                    limit=args.limit,
                    nProc=args.cores,
                    temp_parent=args.temp_parent,
                    retain_temp=args.retain_temp,
                    verbosity=args.verbosity)

    ct.register_files()
    good_chromosomes = ct.run()
    
    _LOGGER.info("Reduce step (merge files)...")
    ct.combine(good_chromosomes)



r"""
## [ROCCO](https://github.com/nolan-h-hamilton/ROCCO/): gwide.py
Runs multiple [`rocco chrom`](https://nolan-h-hamilton.github.io/ROCCO/rocco/chrom.html) jobs,
which execute ROCCO on individual chromosomes. Chromosome-specific parameters are collected in a
CSV file specified with argument `-p --param_file`. See example `--param_file` below for hg38.

If any parameter entry in the CSV file `--param_file` is set to `NULL`, the script uses the corresponding default value
specified with the command-line arguments.

Parameters:
    -p, --param_file (str): Path to the parameter file containing per-chromosome parameters or `hg_params`, `mm_params`
        See hg38_params.csv in the repository for an example CSV file using chromosome-specific budgets.
        `--param_file hg_params` (or `mm_params`) can be used for suggested chromosome-specific parameters.
    -b, --budget (float): Upper-bounds the fraction of loci selected as open.
        Defines the budget constraint (Default: 0.035)
    -g, --gamma (float): $\gamma$ fragmentation penalty in objective func. (Default: 1.0)
    -t, --tau (float): $\tau$ parameter (enrichment threshold) (Default: 0.0).
    --c1 (float): $c_1$ coefficient in the score function (enrichment reward) (Default: 1.0).
    --c2 (float): $c_2$ coefficient in the score function (dispersion penalty) (Default: 1.0).
    --c3 (float): $c_3$ coefficient in the score function (local shift) (Default: 1.0).
    --start (int): Beginning nucleotide position (Default: infer from wiggle files).
    --end (int): Ending nucleotide position (Default: infer from wiggle files).
    -N, --rr_iter (int): Number of RR iterations. $N$ in the paper. (Default: 50).
    --verbose (bool): invoke for verbose logging
    --solver (str): Solver used for the underlying optimization problem of ROCCO (Default: "CLARABEL").
        Other solvers:
            - `MOSEK`: fast, commercial grade solver. Users can instantly obtain a free academic\
                license or generous trial commericial license at
                https://www.mosek.com/products/academic-licenses

            - `PDLP`: first-order solver suitable for very large problems or limited computing environments.
                Uses a primal-dual hybrid gradient approach.

            - `ECOS`: Lightweight second-order conic solver. One of CVXPY's solvers included with the
                default installation, though there are plans to migrate to `CLARABEL`:
                (https://www.cvxpy.org/updates/index.html#ecos-deprecation)
    --bed_format (int): BED format, `3` for BED3 format and `6` for BED6 format (Default: 6).
    --identifiers (str): text file containing a subset of sample IDs, one on each line, for samples\
        to include in the experiment. If not invoked, defaults to `None`, and *all* samples with
        wig files in `--wig_path` are used.
    --outdir (str): Directory in which to store output bed files from the calls to rocco chrom (default: None, create tmp dir.).
    --combine (str): If not None, combine output bed files and store in the file specified with this parameter. For example, `--combine bedname.bed` concatenates the chromosome-specific bedfiles into `bedname.bed` (default: None).
    --multi (int): Run `--multi` rocco chrom jobs simultaneously to improve speed. May increase peak memory use (default: 1).
    --verbose (bool): Set to `True` for verbose logging (default: False).
    --coldata (str): If not None, parse coldata file to create group-specific `--identifiers` files on which to run rocco (default: None).
    --group_column (str): Column in coldata file containing group labels. Only used if --coldata is not None (default: 'group').
    --sample_column (str): Column in coldata file containing sample labels. Only used if --coldata is not None (default: 'sample').
    --split_sex (str): Column in coldata file containing sex labels. Only used if --coldata is not None (default: None).
    --tracks_path (str): location of `tracks_chr[]` subdirectories generated by `rocco prep` if full path not given in `--param_file` 
    --exclude_chroms (str): comma-separated list of chromosomes to ignore

Example `--param_file` for hg38 with chromosome-specific budgets:

    ```
    chromosome,input_path,budget,gamma,tau,c1,c2,c3
    chr1,tracks_chr1,0.035,NULL,NULL,NULL,NULL,NULL
    chr2,tracks_chr2,0.025,NULL,NULL,NULL,NULL,NULL
    chr3,tracks_chr3,0.025,NULL,NULL,NULL,NULL,NULL
    chr4,tracks_chr4,0.02,NULL,NULL,NULL,NULL,NULL
    chr5,tracks_chr5,0.025,NULL,NULL,NULL,NULL,NULL
    chr6,tracks_chr6,0.025,NULL,NULL,NULL,NULL,NULL
    chr7,tracks_chr7,0.03,NULL,NULL,NULL,NULL,NULL
    chr8,tracks_chr8,0.03,NULL,NULL,NULL,NULL,NULL
    chr9,tracks_chr9,0.03,NULL,NULL,NULL,NULL,NULL
    chr10,tracks_chr10,0.025,NULL,NULL,NULL,NULL,NULL
    chr11,tracks_chr11,0.04,NULL,NULL,NULL,NULL,NULL
    chr12,tracks_chr12,0.04,NULL,NULL,NULL,NULL,NULL
    chr13,tracks_chr13,0.025,NULL,NULL,NULL,NULL,NULL
    chr14,tracks_chr14,0.03,NULL,NULL,NULL,NULL,NULL
    chr15,tracks_chr15,0.035,NULL,NULL,NULL,NULL,NULL
    chr16,tracks_chr16,0.035,NULL,NULL,NULL,NULL,NULL
    chr17,tracks_chr17,0.05,NULL,NULL,NULL,NULL,NULL
    chr18,tracks_chr18,0.025,NULL,NULL,NULL,NULL,NULL
    chr19,tracks_chr19,0.05,NULL,NULL,NULL,NULL,NULL
    chr20,tracks_chr20,0.035,NULL,NULL,NULL,NULL,NULL
    chr21,tracks_chr21,0.025,NULL,NULL,NULL,NULL,NULL
    chr22,tracks_chr22,0.35,NULL,NULL,NULL,NULL,NULL
    chrX,tracks_chrX,0.02,NULL,NULL,NULL,NULL,NULL
    chrY,tracks_chrY,0.01,NULL,NULL,NULL,NULL,NULL
    ```

Example `--param_file` for hg38 with the same budget `--budget` for all chromosomes :

    ```
    chromosome,input_path,budget,gamma,tau,c1,c2,c3
    chr1,tracks_chr1,NULL,NULL,NULL,NULL,NULL,NULL
    chr2,tracks_chr2,NULL,NULL,NULL,NULL,NULL,NULL
    chr3,tracks_chr3,NULL,NULL,NULL,NULL,NULL,NULL
    chr4,tracks_chr4,NULL,NULL,NULL,NULL,NULL,NULL
    chr5,tracks_chr5,NULL,NULL,NULL,NULL,NULL,NULL
    chr6,tracks_chr6,NULL,NULL,NULL,NULL,NULL,NULL
    chr7,tracks_chr7,NULL,NULL,NULL,NULL,NULL,NULL
    chr8,tracks_chr8,NULL,NULL,NULL,NULL,NULL,NULL
    chr9,tracks_chr9,NULL,NULL,NULL,NULL,NULL,NULL
    chr10,tracks_chr10,NULL,NULL,NULL,NULL,NULL,NULL
    chr11,tracks_chr11,NULL,NULL,NULL,NULL,NULL,NULL
    chr12,tracks_chr12,NULL,NULL,NULL,NULL,NULL,NULL
    chr13,tracks_chr13,NULL,NULL,NULL,NULL,NULL,NULL
    chr14,tracks_chr14,NULL,NULL,NULL,NULL,NULL,NULL
    chr15,tracks_chr15,NULL,NULL,NULL,NULL,NULL,NULL
    chr16,tracks_chr16,NULL,NULL,NULL,NULL,NULL,NULL
    chr17,tracks_chr17,NULL,NULL,NULL,NULL,NULL,NULL
    chr18,tracks_chr18,NULL,NULL,NULL,NULL,NULL,NULL
    chr19,tracks_chr19,NULL,NULL,NULL,NULL,NULL,NULL
    chr20,tracks_chr20,NULL,NULL,NULL,NULL,NULL,NULL
    chr21,tracks_chr21,NULL,NULL,NULL,NULL,NULL,NULL
    chr22,tracks_chr22,NULL,NULL,NULL,NULL,NULL,NULL
    chrX,tracks_chrX,NULL,NULL,NULL,NULL,NULL,NULL
    chrY,tracks_chrY,NULL,NULL,NULL,NULL,NULL,NULL
    ```

Examples:
    - Run ROCCO genome-wide with default chromosome-specific parameters for hg38 and using all samples for input:
        ```
        rocco gwide --param_file hg_params
        ```
        `hg_params` and `mm_params` correspond to `hg38_params.csv` and `mm10_params.csv` in the ROCCO GitHub repository, respectively.
        Feel free to tweak these templates/default parameters depending on your specific experimental context.

    - Run ROCCO genome-wide with default chromosome-specific parameters for hg38 on a subset of samples listed in `identifiers.txt`:
        ```
        rocco gwide --param_file hg_params --identifiers identifiers.txt
        ```
        For instance, samples are named `sample1, sample2, sample3, ..., sampleK` and `identifiers.txt`
        contains:
        ```
        sample1
        sample2
        sample3
        ```
        Then the command above will ignore sample4,...,sampleK during peak calling.

    - Run ROCCO on human autosomal chromosomes:
        ```
        rocco gwide --param_file hg_params --exclude_chroms chrX,chrY
        ```

    - Run ROCCO on human autosomal chromosomes where the `tracks_chr[]` directories are not in the current working dir.
        ```
        rocco gwide --param_file hg_params --exclude_chroms chrX,chrY --tracks_path /path/to/dir
        ```
    where `dir` contains the subdirectories `tracks_chr1`, `tracks_chr2`, ..., `tracks_chr22` generated by [`rocco prep`](https://github.com/nolan-h-hamilton/ROCCO/).
"""

import os
import sys
import argparse
import subprocess
import tempfile
import pkg_resources
import pandas as pd
import warnings
from . import rocco_aux
from . import chrom

def get_params(param_file: str, budget: float, gamma: float, tau: float,
               c1: float, c2: float, c3: float) -> list:
    """
    Grabs parameters for each chromosome from `param_file`

    This function collects parameters from each chromosome-row of
    `params.csv`. If a `NULL` entry is encountered, it is replaced
    by the CLI-specified default (see `args`).

    Args:
        param_file (str) : a CSV file with a row containing parameter vals\
            to use for each chromosome.

    Returns:
        list: a list of lists, with each element containing
          the necessary parameters to run a `rocco chrom` job.
    """
    defaults = [None, None, budget, gamma, tau, c1, c2, c3]
    chrom_params = []
    with open(param_file, mode='r', encoding='utf-8') as par_file:
        header = True
        for line in par_file:
            if header is True:
                header = False
                continue
            if ',' not in line:
                continue
            line = line.strip()
            line = line.split(',')
            for i, entry in enumerate(line):
                if entry == 'NULL':
                    # replace NULL entries with the genome-wide default
                    line[i] = defaults[i]
            chrom_params.append(line)
    return chrom_params


def call_rocco(chrom, wig_path, budget, gamma, tau, c1, c2, c3, solver,
               bed_format, verbose=False, N=50, identifiers=None,
               outdir='.') -> str:
    r"""
    Formats a command to run `rocco chrom` for a given chromosome

    Args:
        chrom (str): chromosome name. Example: `chr1`
        wig_path (str): path to the *directory* containing samples' signal
            tracks for `chrom`
        budget (float): budget constraint
        tau (float): tau in $\mathcal{S}(i)$
        gam (float): gamma parameter in $f(\mathbf{\ell})$
        c1 (float): c1 value in $\mathcal{S}(i)$
        c2 (float): c2 value in $\mathcal{S}(i)$
        c3 (float): c3 value in $\mathcal{S}(i)$
        N (int): RR procedure iterations. If N <= 0,
            $\texttt{floor\_eps}$ procedure is applied...
            See `Loci.run_rr()`
        verbose_ (bool): Verbosity flag for the solver
        solver (str): the solver to use--either "CLARABEL" or "MOSEK"
    Returns:
        str: a formatted rocco chrom command with the appropriate parameters.
        substituted.
    """
    cli_args = ["rocco", "chrom",
                '--chrom', chrom,
                '--wig_path', wig_path,
                '--budget', budget,
                '--gamma', gamma,
                '--tau', tau,
                '--c1', c1,
                '--c2', c2,
                '--c3', c3,
                '--solver', solver,
                '--bed_format', bed_format,
                '--outdir', outdir,
                '--rr_iter', N,
                '--identifiers', identifiers,
                '--verbose']

    if not verbose:
        cli_args.remove('--verbose')
    if identifiers is None:
        cli_args.remove('--identifiers')
        cli_args.remove(identifiers)
    return ' '.join(cli_args)

def get_params_path(assembly):
    """
    Get the path to the params file for hg38 or mm10
    """
    if 'hg' in assembly:
        return pkg_resources.resource_filename(__name__, 'hg38_params.csv')
    if 'mm' in assembly:
        return pkg_resources.resource_filename(__name__, 'mm10_params.csv')

def main(args):
    default_chrom_spec_params = False
    if args['param_file'] == 'hg_params' and not os.path.exists('hg_params'):
        args['param_file'] = get_params_path('hg38')
        default_chrom_spec_params = True
    if args['param_file'] == 'mm_params' and not os.path.exists('mm_params'):
        args['param_file'] = get_params_path('mm10')
        default_chrom_spec_params = True
    if default_chrom_spec_params and args['tracks_path'] is None:
        print('gwide: `--tracks_path` is `None`...assuming `tracks_chr[]` subdirs are in current directory')

    if args['tracks_path'] is not None:
        args['tracks_path'] = os.path.relpath(args['tracks_path'])

    ignored_chroms = []
    if args['exclude_chroms'] is not None:
        ignored_chroms = args['exclude_chroms'].split(',')

    chrom_args = get_params(args['param_file'], args['budget'],
                            args['gamma'], args['tau'], args['c1'],
                            args['c2'], args['c3'])
    if args['outdir'] is None:
        args['outdir'] = tempfile.mkdtemp(dir=os.getcwd())
        print(f"gwide: --outdir not specified, storing chromosome-specific BED files in {args['outdir']}")
    if args['outdir'] is not None and not os.path.exists(args['outdir']):
        os.mkdir(args['outdir'])

    tmp = tempfile.NamedTemporaryFile(mode="w+")
    for i, arglist in enumerate(chrom_args):
        print(f'\ngwide: chromosome {arglist[0]}')
        arglist = [str(x) for x in arglist]
        if args['tracks_path'] is not None:
            arglist[1] = os.path.join(args['tracks_path'], arglist[1])
        if arglist[0] in ignored_chroms:
            print(f'gwide: skipping --exclude_chroms: {arglist[0]}')
            continue
        if not os.path.exists(arglist[1]) or not len(os.listdir(arglist[1])) > 0:
            print(f'gwide: directory {arglist[1]} does not exist or is empty...skipping')
            continue
        cmd = call_rocco(arglist[0], arglist[1], arglist[2], arglist[3],
                         arglist[4], arglist[5], arglist[6], arglist[7],
                         args['solver'], str(args['bed_format']),
                         args['verbose'], str(args['rr_iter']),
                         identifiers=args['identifiers'], outdir=args['outdir'])

        if args['multi'] == 1:
            try:
                seq_process = subprocess.run(cmd.split(' '),
                                         capture_output=True, text=True, check=True)
                if args['verbose']:
                    print(f'{seq_process.stdout}')
            except Exception as ex:
                raise ex

        tmp.write(str(cmd + '\n'))
    tmp.flush()

    if args['multi'] > 1:
        rocco_aux.run_par(tmp.name, threads=args['multi'], verbose=args['verbose'])
    tmp.close()

    bed_ct = len([x for x in os.listdir(args['outdir']) if x.split('.')[-1] == 'bed'])

    if bed_ct == 0:
        warnings.warn(f"gwide: output directory {args['outdir']} does not contain any bed files")
    elif args['combine'] is not None:
        print('gwide: combining output files --> {}'.format(args['combine']))
        rocco_aux.sort_combine_bed(args['combine'], dir_=args['outdir'])


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--param_file', help='Path to the parameter file containing per-chromosome parameters.\
                        \nSee https://github.com/nolan-h-hamilton/ROCCO/blob/main/rocco/hg38_params.csv for an example with chromosome-specific budgets on hg38.\
                        \n`--param_file hg_params` can be used for suggested chromosome-specific parameters in human samples.\n')
    parser.add_argument('-N', '--rr_iter', type=int, default=50, help = 'number of RR iterations')
    parser.add_argument('--verbose', default=False, action="store_true", help='set to `True` for verbose logging')
    parser.add_argument('-b', '--budget', type=float, default=.035, help='Upper-bounds the fraction of loci selected as open. Defines the budget constraint (Default: 0.035)')
    parser.add_argument('-g', '--gamma', type=float, default=1.0, help='fragmentation penalty in objective func. (Default: 1.0)')
    parser.add_argument('-t', '--tau', type=float, default=0.0, help='tau parameter (enrichment threshold)')
    parser.add_argument('--c1', type=float, default=1.0, help='c_1 coefficient in score function (enrichment reward)')
    parser.add_argument('--c2', type=float, default=1.0, help='c_2 coefficient in score function (dispersion penalty) ')
    parser.add_argument('--c3', type=float, default=1.0, help='c_3 coefficient in score function (local shift)')
    parser.add_argument('--solver', default="CLARABEL",
                        help="Optimization software used to solve the \
                        main LP. `CLARABEL` is used by default.")
    parser.add_argument('--bed_format', type=int, default=3,
                        help="Specifies BED[3] or BED[6] format.")
    parser.add_argument('--identifiers', default=None,
                        help="text file containing a subset of sample IDs, one on each line, for samples to include in the experiment. If not invoked, defaults to `None`, and *all* samples with wig files in `--wig_path` are used. Ignore if using the metadata/coldata parameters.")
    parser.add_argument('--outdir', default=None,
                        help="directory in which to store output bed files from the calls to rocco chrom")
    parser.add_argument('--combine', default=None, help="if not None, combine\
                        output bed files and store in the file specified\
                        with this parameter. ex: `--combine bedname.bed` con-\
                        catenates the chromosome-specific bedfiles into `bedname.bed`.")
    parser.add_argument('--multi', type=int, default=1, help='run `--multi` rocco chrom jobs\
        simultaneously to improve speed. May increase peak memory use.')
    parser.add_argument('--verbose', default=False, action="store_true")
    parser.add_argument('--coldata', default=None, help='if not None, parse coldata file to create group-specific `--identifiers` files on which to run rocco')
    parser.add_argument('--group_column', default='group', help='column in coldata file containing group labels. Only used if --coldata is not None')
    parser.add_argument('--sample_column', default='sample', help='column in coldata file containing sample labels. Only used if --coldata is not None')
    parser.add_argument('--split_sex', default=None, help='column in coldata file containing sex labels. Only used if --coldata is not None')
    parser.add_argument('--tracks_path', default=None, help='location of `tracks_chr[]` subdirectories generated by `rocco prep` if full path not given in `--param_file`')
    parser.add_argument('--exclude_chroms', default=None, help='comma-separated list of chromosomes to ignore. e.g., `--exclude_chroms chr1,chr2,chr5` or `--exclude_chroms chr11`')
    args = vars(parser.parse_args())
    main(args)



"""
## [ROCCO](https://github.com/nolan-h-hamilton/ROCCO/): prep.py 
Obtain ROCCO conformable input from BAM files

BAM files for each sample/replicate in `bamdir` are used to create fixed-step bigwig signal tracks.
These signal tracks are then split by chromosome into wiggle files in `tracks_chr[]`.

The resulting `tracks_chr[]` directories can then be supplied to rocco chrom
via `--wig_path` to construct the matrix $\mathbf{S}_{chr}$.

Parameters:
    -i, --bamdir (str, default='.'):
        Path to the directory containing BAM files.
    -o, --outdir (str, default='.')`:
        Output directory.
    -s, --sizes (str, default='hg38'):
        A path to a chromosome sizes file OR an assembly name.
    -L, --interval_length (int, default=50)`:
        Wiggle track fixed step size
    -c, --cores (int, default=1)`:
        The `bamSitesToWig.py` cores parameter. Altering this parameter to use >1 core may cause issues on Mac OS.

Examples:
    - create input `tracks_chr[]` directories for the hg38 assembly, no sizes file available:
        ```
        rocco prep -s hg38 --bamdir [/path/to/bamfiles]
        ```
    - create input `tracks_chr[]` directories from BAM files using a locally available sizes file:
        ```
        rocco prep -s [/path/to/sizefile] --bamdir [path/to/bamfiles]
        ```
    - create input`tracks_chr[]` directories from hg19 BAM files in the current directory:
        ```
        rocco prep -s hg19 --bamdir .
        ```
"""

import os
import argparse
import subprocess
import pysam
from . import rocco_aux
import tempfile


def main(args):
    # process command line arguments
    cwd = os.getcwd()
    args['bamdir'] = os.path.abspath(args['bamdir'])
    args['outdir'] = os.path.abspath(args['outdir'])

    if not os.path.exists(args['sizes']):
        # if args['sizes'] is not a locally available file,
        # check UCSC for the assembly's sizes file
        try:
            downloaded_size_file = rocco_aux.get_size_file(args['sizes'])
        except Exception as ex:
            print(f'prep: {args["sizes"]} could not be found locally as a file (e.g. `hg38.sizes`) or on UCSC server as an assembly name (e.g. `hg38`).')
            raise
        args['sizes'] = downloaded_size_file

    chroms = list(rocco_aux.parse_size_file(args['sizes']).keys())
    bigwig_files = []
    tf = tempfile.NamedTemporaryFile(mode='w', delete=False)

    for fname in os.listdir(args['bamdir']):
        fname = os.path.join(args['bamdir'], fname)
        fname = os.path.abspath(fname)

        # create an index for bam files if not present already
        if rocco_aux.is_alignment(fname):
            aln = pysam.AlignmentFile(fname)
            try:
                aln.check_index()
            except ValueError:
                print(f'prep: no index file available for {fname}, calling pysam.index()')
                pysam.index(fname)

            print('prep: {}'.format(aln.filename.decode()))
            bstw_path = os.path.join(os.path.dirname(__file__), "bamSitesToWig.py")
            bstw_cmd = ['python3', bstw_path,
                        '-i', aln.filename.decode(),
                        '-c', args['sizes'],
                        '-w', args['outdir'] + '/' + aln.filename.decode().split('/')[-1] + '.bw',
                        '-r', str(args['interval_length']),
                        '-m', 'atac',
                        '-l', str(int(args['interval_length'])//2),
                        '-p', str(args['cores']),
                        '--variable-step',
                        '--limit', " ".join([x for x in chroms if rocco_aux.has_reads(bamfile=aln.filename.decode(),min_reads=1,chrom=x)])]

            if not args['multi']:
                # if `--multi False
                subprocess.run(bstw_cmd,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL,
                               shell=False, check=True)
            else:
                tf.write(' '.join(bstw_cmd))
                tf.write('\n')

            bigwig_file = args['outdir'] + '/' + aln.filename.decode().split('/')[-1] + '.bw'
            bigwig_files.append(bigwig_file)
    tf.close()

    if args['multi']:
        rocco_aux.run_par(tf.name)

    # after running bamSitesToWig, split each bigwig file by chromosome into tracks_chr[]
    os.chdir(args['outdir'])
    for chrom in chroms:
        try:
            os.mkdir(f'tracks_{chrom}')
        except FileExistsError:
            pass
        for bigwig_file in [bw_file.split('/')[-1] for bw_file in bigwig_files]:
            subprocess.run(['bigWigToWig', bigwig_file,
                        chrom + '_' + bigwig_file + '.wig',
                        '-chrom={}'.format(chrom)], check=True)

            # if wig file is nonempty, place in tracks_chr[]
            if os.path.getsize(chrom + '_' + bigwig_file + '.wig') > 1024:
                subprocess.run(["mv", chrom + '_' + bigwig_file + '.wig',
                        "tracks_{}".format(chrom)], check=True)
            else:
                subprocess.run(["rm", chrom + '_' + bigwig_file + '.wig'], check=True)

    os.chdir(cwd)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='BAM preprocessing')
    parser.add_argument('-i','--bamdir', default='.', type=str, help='Path to the directory containing BAM files.')
    parser.add_argument('-o', '--outdir', default= '.', help='Output directory')
    parser.add_argument('-s', '--sizes',
                        default='hg38',
                        help="A path to a chromosome sizes file. OR\
                        an assembly name")
    parser.add_argument('-L', '--interval_length',
                        default=50,
                        help="wiggle track fixed step size")
    parser.add_argument('-c', '--cores',
                        type=int,
                        default=1,
                        help="`bamSitesToWig.py`'s cores  parameter.\
                        Altering this parameter to use >1 core\
                        may cause issues on Mac OS")
    parser.add_argument('--multi',
                         default=True)
    args = vars(parser.parse_args())
    main(args)



"""
## [ROCCO](https://github.com/nolan-h-hamilton/ROCCO/): locus.py 
"""
class Locus:
    r"""
    `Locus` objects are the $n \approx \frac{|\mathscr{L}|}{L}$ nodes
    dividing genomic region $\mathscr{L}$ into segments of $L$ nucleotides.

    Attributes:
        position (int): Nucleotide position of Locus `self`.
        left (Locus): Points to the previous adjacent locus.
        right (Locus): Points to the subsequent adjacent locus.
        size (int): Size (in nucleotides) of the Locus `self`.
            Corresponds to $L$ before merging.
        parent (Locus): Parent locus of the Locus `self`.
            Only defined for Locus objects merged by `Loci.combine()`.
        subloci (list): List of subloci comprising Locus objet `self`.
            Only defined for Locus objects that were created from merged
            loci with `Loci.combine()`.
        sig_data (np.ndarray): $K \times 1$ vector of replicates' signal
            values at Locus `self`. Corresponds to $\mathbf{s}_i$.
        accessible (float): a value between 0 and 1. Corresponds to
            $\ell_i$.

    Note:
        After optimization, ROCCO merges contiguous, accessible Locus objects.
            Meaning `Locus.size` may be larger than the initial locus size, and
            is in this way inconsistent with the definition of $L$ in the paper.
    """
    def __init__(self, position=None,left=None,right=None,
                 size=None, parent=None, subloci=None,
                 sig_data=None, accessible=None):
        self.position = position
        self.left = left
        self.right = right
        self.size = size
        self.parent = parent
        self.subloci = subloci
        self.sig_data = sig_data
        if accessible is None:
            self.accessible = -1
        else:
            self.accessible = accessible





r"""
# ROCCO: [R]obust [O]pen [C]hromatin Detection via [C]onvex [O]ptimization.
Underlying ROCCO is a constrained optimization problem that can be solved efficiently to predict consensus regions of open chromatin across many samples.

#### GitHub (Homepage): https://github.com/nolan-h-hamilton/ROCCO/

#### Demo: https://github.com/nolan-h-hamilton/ROCCO/blob/main/demo/demo.ipynb

#### Paper: https://doi.org/10.1101/2023.05.24.542132


Subcommand Documentation:
    [`gwide`](https://nolan-h-hamilton.github.io/ROCCO/rocco/gwide.html)

    [`chrom`](https://nolan-h-hamilton.github.io/ROCCO/rocco/chrom.html)

    [`prep`](https://nolan-h-hamilton.github.io/ROCCO/rocco/prep.html)

    [`budgets`](https://nolan-h-hamilton.github.io/ROCCO/rocco/budgets.html)

General Notation and Terminology:
    - $\mathscr{L}$: genomic region, e.g., a chromosome
    - $L$: size of loci in bp, i.e., fixed step size in wiggle tracks
    - $n$: number of loci in the given chromosome, defined by $n \approx |\mathscr{L}|/L$.
    - $\mathcal{S}(i)$: score for $i$th locus:
        $$c_1 g_1(i) - c_2 g_2(i) + c_3 g_3(i)$$
    - $\ell \in [0,1]^n$ (relaxed) or $\ell \in \mathbb{Z}^n_{0,1}$ (unrelaxed): solutions to the optimization problem, $\ell$, specify which loci are accessible ($\ell_i = 1$) and which are closed ($\ell_i = 0$)
    - objective function $f$: $f(\mathbf{\ell}) = -\mathcal{S}^{T}\mathbf{\ell} + \gamma \sum_{i=1}^{n-1}|\ell_i - \ell_{i+1}|$
        - the first term represents the sum of scores, $\sum_i \mathcal{S}(i)$, over selected loci
        - the second term induces sparsity and controls fragmentation of selected regions
    - budget constraint: $\sum_{i=1}^{n}\ell_i \leq [nb]$. Upper bounds the proportion of loci selected as accessible.
    - Relaxed optimization problem to obtain initial solution:
        $$\text{Minimize: } \ell \in [0,1]^n, f(\mathbf{\ell}) = -\mathcal{S}^{T}\mathbf{\ell} + \gamma \sum_{i=1}^{n-1}|\ell_i - \ell_{i+1}|$$
        $$\text{Subject to: }\sum_{i=1}^{n}\ell_i \leq [nb]$$
    - `RR`: iterative randomization procedure described in the paper to derive integral solutions (open chromatin annotations) from the relaxed solution as:
        $$\ell^{\textsf{rand}}_i \sim \text{Bernoulli}(\ell_i)$$
        `--rr_iter` such solutions are generated, after which the best feasible solution is selected to determine the final annotation.

"""

#!/usr/bin/env python
import os
import argparse
import subprocess
from . import rocco_aux
from . import chrom
from . import gwide
from . import prep
from . import budgets
from . import locus
from . import loci
import pandas as pd

def parse_coldata(coldata_file: str, group_column: str, sample_column: str, split_sex: str = None, delimiter='\t'):
    """
    Parses a metadata file for the samples. Useful if running on subgroups separately.

    Args:
        coldata_file (str): path to the colData file in CSV format
        group_column (str):  column containing group label
        sample_column (str): column containing sample names
        split_sex (str): if not None, create separate text files for each sex within each group
            based on the specified column name.
        delimiter (str): delimiter used in coldata file

    Returns:
        List: a list of the file names created
    """
    created_files = []
    try:
        coldata_df = pd.read_csv(coldata_file, delimiter=delimiter)
        group_names = coldata_df[group_column].unique()
        for group_name in group_names:
            group_samples = coldata_df[coldata_df[group_column] == group_name][sample_column]
            file_name = f'group_{group_name}'
            if split_sex is None:
                print(f'rocco.parse_coldata(): writing {file_name}')
                with open(file_name, 'w') as file:
                    file.write('\n'.join(group_samples))
                    file.write('\n')
                created_files.append(file_name)
            if split_sex is not None:
                for sex in coldata_df[split_sex].unique():
                    sex_samples = coldata_df[
                        (coldata_df[group_column] == group_name) &
                        (coldata_df[split_sex] == sex)
                    ][sample_column]
                    sex_file_name = f'group_{group_name}_sex_{sex}'
                    print(f'rocco.parse_coldata(): writing {sex_file_name}')
                    with open(sex_file_name, 'w') as file:
                        file.write('\n'.join(sex_samples))
                        file.write('\n')
                    created_files.append(sex_file_name)
    except Exception as e:
        print(f"rocco.parse_coldata(): Parsing coldata failed. Ensure column names are correctly specified.")
        raise
    return created_files

def subcommand_chrom(args):
    chrom.main(args)
    return True

def subcommand_gwide(args):
    if args['coldata'] is not None:
        ind_files = parse_coldata(args['coldata'],
                                  args['group_column'],
                                  args['sample_column'],
                                  args['split_sex'])
        bed_files = []
        for file_ in ind_files:
            temp_args = args.copy()
            print(f'rocco gwide: {file_}')
            temp_args['identifiers'] = file_
            temp_args['outdir'] = file_ + '_dir'
            temp_args['combine'] = file_ + '.bed'
            bed_files.append(temp_args['combine'])
            gwide.main(temp_args)

        print(f'rocco gwide completed: {bed_files}')
    else:
        gwide.main(args)

    return True

def subcommand_prep(args):
    prep.main(args)
    return True

def subcommand_budgets(args):
    budgets.main(args)
    return True


def main():
    parser = argparse.ArgumentParser(description="ROCCO: [R]obust [O]pen [C]hromatin Dection via [C]onvex [O]ptimization.\n\nPyPI : https://pypi.org/project/rocco/ \nGitHub: https://github.com/nolan-h-hamilton/ROCCO/ \nDemo: https://github.com/nolan-h-hamilton/ROCCO/blob/main/demo/demo.ipynb \nPaper: https://doi.org/10.1101/2023.05.24.542132\n", add_help=True, formatter_class=argparse.RawTextHelpFormatter)
    subparsers = parser.add_subparsers(dest="command")

    # 'gwide' subcommand parameters
    parser_subcommand_gwide = subparsers.add_parser("gwide", help='run rocco genome-wide/on multiple chromosomes (gwide.py)')
    parser_subcommand_gwide.add_argument('-p', '--param_file', help='Path to the parameter file containing per-chromosome parameters.\
                        \nSee https://github.com/nolan-h-hamilton/ROCCO/blob/main/rocco/hg38_params.csv for an example with chromosome-specific budgets on hg38.\
                        \n`--param_file hg_params` can be used for suggested chromosome-specific parameters in human samples.\n')
    parser_subcommand_gwide.add_argument('-b', '--budget', type=float, default=.035)
    parser_subcommand_gwide.add_argument('-g', '--gamma', type=float, default=1.0)
    parser_subcommand_gwide.add_argument('-t', '--tau', type=float, default=0.0)
    parser_subcommand_gwide.add_argument('--c1', type=float, default=1.0)
    parser_subcommand_gwide.add_argument('--c2', type=float, default=1.0)
    parser_subcommand_gwide.add_argument('--c3', type=float, default=1.0)
    parser_subcommand_gwide.add_argument('-N', '--rr_iter', type=int, default=50)
    parser_subcommand_gwide.add_argument('--solver', type=str, default="CLARABEL")
    parser_subcommand_gwide.add_argument('--bed_format', type=int, default=3)
    parser_subcommand_gwide.add_argument('--identifiers', default=None)
    parser_subcommand_gwide.add_argument('--outdir', default=None, help='if `None`, a temporary directory will be created')
    parser_subcommand_gwide.add_argument('--combine', default=None)
    parser_subcommand_gwide.add_argument('--multi', default=1, type=int, help='number of simultaneous `rocco chrom` jobs to execute')
    parser_subcommand_gwide.add_argument('--verbose', default=False, action="store_true")
    parser_subcommand_gwide.add_argument('--coldata', default=None, help='if not None, parse coldata file to create group-specific `--identifiers` files on which to run rocco')
    parser_subcommand_gwide.add_argument('--group_column', default='group', help='column in coldata file containing group labels. Only used if --coldata is not None')
    parser_subcommand_gwide.add_argument('--sample_column', default='sample', help='column in coldata file containing sample labels. Only used if --coldata is not None')
    parser_subcommand_gwide.add_argument('--split_sex', default=None, help='column in coldata file containing sex labels. Only used if --coldata is not None')
    parser_subcommand_gwide.add_argument('--tracks_path', default=None)
    parser_subcommand_gwide.add_argument('--exclude_chroms', default=None)
    # 'chrom' subcommand parameters
    parser_subcommand_chrom = subparsers.add_parser("chrom", help='run ROCCO on a single chromosome (chrom.py)')
    parser_subcommand_chrom.add_argument('--start', type=int, default=-1)
    parser_subcommand_chrom.add_argument('--end', type=int, default=-1)
    parser_subcommand_chrom.add_argument('--locus_size', type=int, default=-1)
    parser_subcommand_chrom.add_argument('-b', '--budget', type=float, default=.035)
    parser_subcommand_chrom.add_argument('--chrom', type=str, help='')
    parser_subcommand_chrom.add_argument('--wig_path', type=str, default=os.getcwd())
    parser_subcommand_chrom.add_argument('-g', '--gamma', type=float, default=1.0)
    parser_subcommand_chrom.add_argument('-t', '--tau', type=float, default=0.0)
    parser_subcommand_chrom.add_argument('--c1', type=float, default=1.0)
    parser_subcommand_chrom.add_argument('--c2', type=float, default=1.0)
    parser_subcommand_chrom.add_argument('--c3', type=float, default=1.0)
    parser_subcommand_chrom.add_argument('--solver', type=str,default="CLARABEL")
    parser_subcommand_chrom.add_argument('--bed_format', type=int, default=6)
    parser_subcommand_chrom.add_argument('--identifiers', default=None)
    parser_subcommand_chrom.add_argument('--outdir', type=str, default='.')
    parser_subcommand_chrom.add_argument('-N', '--rr_iter', type=int, default=50)
    parser_subcommand_chrom.add_argument('--verbose', default=False, action="store_true")

    # 'prep' subcommand parameters
    parser_subcommand_prep = subparsers.add_parser("prep", help='Preprocess BAM files (prep.py)')
    parser_subcommand_prep.add_argument('-i', '--bamdir', default='.', type=str)
    parser_subcommand_prep.add_argument('-o', '--outdir', default='.')
    parser_subcommand_prep.add_argument('-s', '--sizes', default='hg38')
    parser_subcommand_prep.add_argument('-L', '--interval_length', default=50)
    parser_subcommand_prep.add_argument('-c', '--cores', type=int, default=1)
    parser_subcommand_prep.add_argument('--multi', default=True)
    parser_subcommand_prep.add_argument('--bstw_path', default=None, help='deprecated')

    # 'budgets' subcommand parameters
    parser_subcommand_budgets = subparsers.add_parser("budgets", help='Compute a budget for each chromosome ranked by read density (budgets.py). ')
    parser_subcommand_budgets.add_argument('--wigdir', type=str)
    parser_subcommand_budgets.add_argument('-s', '--sizes', type=str)
    parser_subcommand_budgets.add_argument('--smean', type=float, default=.035)
    parser_subcommand_budgets.add_argument('--samp_rate', type=float, default=0.10)
    parser_subcommand_budgets.add_argument('-o', '--outfile', type=str, default='params.csv')
    parser_subcommand_budgets.add_argument('-L','--locsize',type=int,default=50)

    # 'get_sizes' subcommand parameters
    parser_subcommand_get_sizes = subparsers.add_parser("get_sizes", help="download sizes file for a genome in the ucsc genome registry")
    parser_subcommand_get_sizes.add_argument('-g', '--genome', type=str, default='hg38')

    args = vars(parser.parse_args())
    if args['command'] == "gwide":
        subcommand_gwide(args)
    elif args['command'] == "chrom":
        subcommand_chrom(args)
    elif args['command'] == 'prep':
        subcommand_prep(args)
    elif args['command'] == 'budgets':
        subcommand_budgets(args)
    elif args['command'] == 'get_sizes':
        rocco_aux.get_size_file(args['genome'])
    else:
        parser.print_help()

if __name__ == "__main__":
    main()



r"""
## [ROCCO](https://github.com/nolan-h-hamilton/ROCCO/): loci.py 
ROCCO creates a Locus object (see `Locus.py`) for each
$\ell_i, \forall i = 1 \ldots n$. These Locus objects are then linked
together as a Loci object, defined in this code.

Scoring: `Loci.score_loci()`

Optimization: `Loci.rocco_lp()`

Integral solutions: `Loci.run_rr()`

"""

import copy
import math
import random
import numpy as np
import cvxpy as cp
from scipy import stats
from .locus import Locus

class LociIterator:
    """
    Iterator for traversing the Loci object

    Args:
        head (Locus): first Locus object in the Loci object

    Returns:
        LociIterator: an iterator object

    Raises:
        StopIteration: raised if end of genomic region is reached during iteration
    """
    def __init__(self, head):
        self.current = head

    def __iter__(self):
        return self

    def __next__(self):
        if not self.current:
            raise StopIteration

        item = copy.copy(self.current)
        self.current = self.current.right
        return item

class Loci:

    def __init__(self, head=None, tail=None):
        self.head = head
        self.tail = tail

    def __iter__(self):
        return LociIterator(self.head)

    def is_empty(self) -> bool:
        """
        Check if the Loci object is empty

        Returns:
            bool: True if Loci object is empty--False otherwise.
        """
        if self.head is None and self.tail is None:
            return True
        return False

    def append_locus(self,locus):
        """
        Append a new locus as the tail of the Loci object.

        Args:
            locus (Locus): The locus to be appended.

        Returns:
            None
        """
        if  self.is_empty():
            self.head = locus
            self.tail = locus
            locus.left=None
            locus.right=None
        else:
            tail_cpy = self.tail
            self.tail = locus
            locus.right = None
            locus.left = tail_cpy
            tail_cpy.right=locus

    def combine(self,head_,tail_):
        """Combine contiguous accessible regions

        This function takes the starting and ending node
        for a sequence of contiguous loci (i.e., a sequence
        such that all loci satisfy `Locus.accessible = 1`) and
        merges them into a single locus object. It defines
        the new locus's signal vector as the sum
        of signal vectors among loci in the contiguous
        sequence.

        Args:
            head_ (Locus): first locus object in sequence of contiguous
              selected loci
            tail_ (Locus): last locus object in sequence of contiguous
              selected loci
        """
        cpy = head_
        accessible_ = 0
        size_ = 0
        sig_matrix = []
        subloc = Loci()

        accessible_= 0
        if head_.accessible == 1:
            accessible_ = 1

        while head_ is not None and head_.position <= tail_.position:
            size_ += head_.size
            sig_matrix.append(head_.sig_data)
            subloc.append_locus(copy.copy(head_))
            head_ = head_.right

        sig_matrix = np.array([np.array(sig) for sig in sig_matrix]).transpose()
        sig_data_ = np.sum(sig_matrix,axis=1)

        # create new locus
        new_loc = Locus(position=cpy.position,left=cpy.left,right=head_,
                        size=size_, parent=None, subloci=subloc,
                        sig_data=sig_data_,accessible=accessible_)

        if head_ is None:
            self.tail = new_loc
        else:
            head_.left = new_loc
            cpy.right = new_loc

        if cpy.left is not None:
            cpy.left.right = new_loc
        else:
            self.head = new_loc
        s_head_ = subloc.head
        while s_head_ is not None and s_head_.position <= tail_.position:
            s_head_.parent=new_loc
            s_head_ = s_head_.right

    def as_list(self) -> list:
        """
        Get a list representation of the Loci object.

        Returns:
            list: list representation of the Loci object, where the Locus elements
                are ordered as in the Loci object.
        """
        head_ = self.head
        list_repr = []
        while head_ is not None:
            list_repr.append(copy.copy(head_))
            head_ = head_.right
        return list_repr

    def combine_selected(self):
        r"""
        Find adjacent accessible regions and combine them into a single locus

        This function identifies sequences of locus objects, $\ell_i$, satisfying
        $\ell_i$`.accessible > 0` and combines them into a single locus via `Loci.combine()`.


        Notes:
            - `Loci.combine()` defines signal data for the created locus as the vector sum of
                signal data for each Locus object comprising the contiguous accessible region.

        Returns:
            None
        """
        head_ = self.head
        while head_ is not None:
            to_combine = []
            head_cpy = head_
            while head_cpy is not None and head_cpy.accessible > 0 :
                to_combine.append(head_cpy)
                head_cpy = head_cpy.right
            if len(to_combine) > 1:
                self.combine(to_combine[0], to_combine[-1])
            head_ = head_cpy
            if head_ is not None:
                head_ = head_.right

    def get_sig_mat(self) -> np.ndarray:
        r"""
        Builds the signal matrix $\mathbf{S}_{chr}$ as defined in the paper.

        Each Locus object's `sig_data` attribute is a $K \times 1$ column
        vector, so $\mathbf{S}_{chr}$ is a $K \times n$ matrix.

        Returns:
            None
        """
        sig_mat = []
        head_ = self.head
        while head_ is not None:
            sig_mat.append(head_.sig_data)
            head_ = head_.right
        return np.array(sig_mat)

    def score_loci(self, tau: float = 0.0, c1: float = 1.0, c2: float = 1.0, c3: float = 1.0, eps_l=1e-4) -> np.ndarray:
        r"""
        compute locus score vector, i.e., $\mathcal{S}(i), \forall i = 1 \ldots n$
        given the $K \times n$ signal matrix $\mathbf{S}_{chr}$.

        Args:
            tau (float): median signal threshold (see $\tau$ in paper)
            c1 (float): weight for the first term in  $\mathcal{S}(i)$,
                (median signal) $g_1$
            c2 (float): weight for second term in $\mathcal{S}(i)$,
                (MAD dispersion) $g_2$
            c3 (float): weight for third term in  $\mathcal{S}(i)$, $g_3$
            eps_l (float): prevent unnecessary selections in interior-point solutions

        Returns:
            np.ndarray: the $n$-dimensional score vector $\mathcal{S}(i), \forall i = 1 \ldots n$
        """
        def g3_vec(vvec) -> np.ndarray:
            g3_vals = np.zeros(len(vvec))
            for i,Loc in enumerate(vvec):
                if i == 0:
                    vel = (abs(vvec[i] - vvec[i+1]))
                if i == len(vvec)-1:
                    vel = abs(vvec[i] - vvec[i-1])
                else:
                    vel = max(abs(vvec[i] - vvec[i+1]),
                        abs(vvec[i] - vvec[i-1]))

                vel /= vvec[i]+1
                g3_vals[i] = vel
            return g3_vals

        sig_mat = self.get_sig_mat()
        med_vec = np.median(sig_mat,axis=1) # g_1
        mad_vec = stats.median_abs_deviation(sig_mat,axis=1) # g_2
        g3_vec_ = g3_vec(med_vec) # g_3
        s_vec = c1*med_vec - c2*mad_vec + c3*g3_vec_

        for i, val in enumerate(s_vec):
            if med_vec[i] <= tau:
                s_vec[i] = 0
        return s_vec - eps_l

    def run_rr(self, lp_sol, N, loci_scores, budget, gam, eps = 1e-8) -> np.ndarray:
        r"""
        Carry out the $\texttt{RR}$ rounding procedure to return a
        'good' integral solution with reference to the LP solution

        Args:
            lp_sol (numpy.ndarray): The LP solution
            N (int): Number of RR solutions to evaluate. If $N \leq 0$, the
                floor-epsilon protocol is applied.
            loci_scores (np.ndarray): $\mathcal{S}$
            budget (float): budget parameter
            gam (float): weight for $\sum_{i=1}^{i=n-1} |\ell_i - \ell_{i+1}|$ (discontig.) penalty
            eps (float): `init_sol = np.floor(lp_sol + eps)`. Decreased iteratively if initial `eps` does
                not yield a feasible solution.

        Returns:
        np.ndarray: the integral $\texttt{RR}$ solution, OR $\texttt{floor\_eps}$ solution if $N \leq 0$.
    """
        n = len(loci_scores)
        eps_cpy = eps
        # initialize as floor_eps solution
        init_sol = np.floor(lp_sol + eps)
        while np.sum(init_sol) > np.floor(n*budget):
            # loop guarantees the feasibility of solutions
            eps_cpy = eps_cpy/2
            init_sol = np.floor(lp_sol + eps_cpy)
        if N <= 0:
            return init_sol

        init_score = (-loci_scores@init_sol
                       + gam*np.sum(np.abs(np.diff(init_sol,1))))

        nonint_loci = [i for i in range(len(lp_sol)) if lp_sol[i] > 0 and lp_sol[i] < 1]
        rr_sol = init_sol
        best_score = init_score
        for j in range(N):
            ell_rand_n = copy.copy(lp_sol)
            # for efficiency, only round `nonint_loci`
            for idx in nonint_loci:
                if random.random() <= lp_sol[idx]:
                    ell_rand_n[idx] = 1
                else:
                    ell_rand_n[idx] = 0

            score = (-loci_scores@ell_rand_n
                       + gam*np.sum(np.abs(np.diff(ell_rand_n,1))))

            is_feas = (np.sum(ell_rand_n) <= math.floor(n*budget))
            if is_feas and score < best_score:
                rr_sol = ell_rand_n
                best_score = score
        return rr_sol

    def rocco_lp(self, budget: float = .035, tau: float = 0, gam: float = 1, c1: float = 1,
             c2: float = 1, c3: float = 1, verbose_: bool = False, solver: str = "CLARABEL",
             N: int = 50, solver_reltol: float = 1.0e-8, solver_maxiter: float = 10000,
             solver_feastol: float = 1e-8) -> cp.Problem:
        r"""
        Solve the LP-relaxation and apply the randomization procedure,
        $\texttt{RR}$ ($N > 0$), or $\texttt{floor\_eps}$ ($N \leq 0$) for an
        integral solution $\mathbf{\ell}^{*}$.

        Args:
            budget (float): determines budget constraint $\sum_i \ell_i \leq nb$
            tau (float): $\tau$ in $\mathcal{S}$
            gam (float): $\gamma$ in obj. function $f(\mathbf{\ell})$
            c1 (float): $c_1$ value in $\mathcal{S}$.
            c2 (float): $c_2$ value in $\mathcal{S}$
            c3 (float): $c_3$ value in $\mathcal{S}$
            N (int): $\texttt{RR}$ procedure iterations. See `Loci.run_rr()`
            verbose_ (bool): Verbosity flag for the solver
            solver (str): Defaults to `CLARABEL`. `ECOS, PDLP, MOSEK` are also supported.
            solver_reltol (float): acceptable relative optimality gap for solver termination.
            solver_feastol (float): the solver will consider solutions within `solver_feastol`
                of the constraint boundaries as feasible
            solver_maxiter (int): bounds the number of solving iterations before terminating
        Returns:
            cp.Problem: a CVXPY problem object
    """
        loci_scores = self.score_loci(tau, c1, c2, c3) # S_i, i=1...n
        n = len(loci_scores)
        # define problem in CVXPY
        ell = cp.Variable(n)
        z = cp.Variable(n-1) # aux. variables
        constraints = [ell <= 1,
                          ell >= 0,
                          cp.sum(ell) <= math.floor(budget*n),
                          z >= cp.diff(ell,1),
                          z >= -1*cp.diff(ell,1)]
        problem = cp.Problem(cp.Minimize(-loci_scores@ell + gam*cp.sum(z)),
                             constraints)

        # Refer to `Solve method options` at
        # https://www.cvxpy.org/tutorial/advanced/index.html
        if solver.lower() not in [x.lower() for x in cp.installed_solvers()]:
            raise cp.error.SolverError(
                f'\nSolver {solver} is not available.\
                \nInstalled solvers: {cp.installed_solvers()}\
                \nhttps://www.cvxpy.org/tutorial/advanced/index.html\n')

        elif solver.lower() == 'clarabel':
            problem.solve(solver=cp.CLARABEL,
                          tol_gap_rel=solver_reltol,
                          tol_feas=solver_feastol,
                          max_iter=solver_maxiter,
                          verbose=verbose_)

        elif solver.lower() == "ecos":
            problem.solve(solver=cp.ECOS,
                          reltol=solver_reltol,
                          max_iters=solver_maxiter,
                          feastol=solver_feastol,
                          verbose=verbose_)

        elif solver.lower() == "mosek":
            # https://docs.mosek.com/latest/pythonapi/parameters.html
            MOSEK_OPTS = {'MSK_IPAR_NUM_THREADS': 1,
                        'MSK_IPAR_PRESOLVE_LINDEP_ABS_WORK_TRH': 10}
            try:
                problem.solve(cp.MOSEK, mosek_params=MOSEK_OPTS, eps=solver_reltol, bfs=True, verbose=verbose_)
            except cp.error.SolverError as ex:
                print("Ensure a valid MOSEK license is available: https://docs.mosek.com/latest/licensing/quickstart.html")
                raise ex

        elif solver.lower() == "pdlp":
            problem.solve(solver=cp.PDLP, verbose=verbose_)

        p_stat = problem.status
        if p_stat is None or p_stat in ['infeasible','unbounded'] or problem.variables()[0].value is None:
            raise cp.error.SolverError(f'\nFailed to obtain optimal solution.\
                \nProblem status: {problem.status}.\n')

        lp_sol = problem.variables()[0].value
        sol_rr  = self.run_rr(lp_sol, N, loci_scores, budget, gam)

        head_ = self.head
        i = 0
        while head_ is not None:
            head_.accessible = sol_rr[i]
            head_ = head_.right
            i += 1

        return problem



r"""
## [ROCCO](https://github.com/nolan-h-hamilton/ROCCO/): chrom.py 
Run ROCCO on chromosome `--chrom` with samples' count/signal data in `tracks_chr[]`

This script is the workhorse for [`rocco gwide`](https://nolan-h-hamilton.github.io/ROCCO/rocco/gwide.html) which generates and combines
results for multiple chromosomes

Parameters:
    --chrom (str): (Required) Specifies the chromosome, e.g., `--chrom chr1`.
    --wig_path (str): (Required) Directory generated by `rocco prep` containing chromosome-specific
        .wig files for each sample, e.g., `--wig_path tracks_chr1`
    -b, --budget (float): Upper-bounds the fraction of loci selected as open.
        Defines the budget constraint (Default: 0.035)
    -g, --gamma (float): $\gamma$ fragmentation penalty in objective func. (Default: 1.0)
    -t, --tau (float): $\tau$ parameter (enrichment threshold) (Default: 0.0).
    --c1 (float): $c_1$ coefficient in the score function (enrichment reward) (Default: 1.0).
    --c2 (float): $c_2$ coefficient in the score function (dispersion penalty) (Default: 1.0).
    --c3 (float): $c_3$ coefficient in the score function (local shift) (Default: 1.0).
    --start (int): Beginning nucleotide position (Default: infer from wiggle files).
    --end (int): Ending nucleotide position (Default: infer from wiggle files).
    -N, --rr_iter (int): Number of RR iterations. $N$ in the paper. (Default: 50).
    --verbose (bool): invoke for verbose logging
    --solver (str): Solver used for the underlying optimization problem of ROCCO (Default: "CLARABEL").
        Other solvers:
            - `MOSEK`: fast, commercial grade solver. Users can instantly obtain a free academic\
                license or generous trial commericial license at
                https://www.mosek.com/products/academic-licenses

            - `PDLP`: first-order solver suitable for very large problems or limited computing environments.
                Uses a primal-dual hybrid gradient approach.

            - `ECOS`: Lightweight second-order conic solver. One of CVXPY's solvers included with the
                default installation, though there are plans to migrate to `CLARABEL`:
                (https://www.cvxpy.org/updates/index.html#ecos-deprecation)
    --bed_format (int): BED format, `3` for BED3 format and `6` for BED6 format (Default: 6).
    --identifiers (str): text file containing a subset of sample IDs, one on each line, for samples\
        to include in the experiment. If not invoked, defaults to `None`, and *all* samples with
        wig files in `--wig_path` are used.

Input:
    A `tracks_chr[]` directory containing samples wiggle tracks created by [`rocco prep`](https://nolan-h-hamilton.github.io/ROCCO/rocco/prep.html).
    For example: `--wig_path tracks_chr1`

Output:
    BED file containing predicted regions of open chromatin. Named with convention:
    ```ROCCO_out_[chrom]_[budget]_[gamma]_[tau]_[c1]_[c2]_[c3].bed```

Examples:
    - Run ROCCO on chromosome 21 with default parameters and using all samples for input:
        ```
        rocco chrom --wig_path tracks_chr21 --chrom chr21
        ```
    - Run ROCCO on chromosome 21 with default parameters on a subset of samples listed in `identifiers.txt`:
        ```
        rocco chrom --wig_path tracks_chr21 --chrom chr21 --identifiers identifiers.txt
        ```
        For instance, samples are named `sample1, sample2, sample3, ..., sampleK` and `identifiers.txt`
        contains:
        ```
        sample1
        sample2
        sample3
        ```
        Then the command above will ignore sample4,...,sampleK during peak calling.

    - Run ROCCO on chromosome 21. Select no more than $2\%$ of the $n$ loci
        ```
        rocco chrom --wig_path tracks_chr21 --chrom chr21 --budget 0.02
        ```
"""

import os
import sys
import gc
import argparse
import warnings
import subprocess
import numpy as np
import pandas as pd
from .locus import Locus
from .loci import Loci


def get_locus_size(wig_path) -> int:
    """
    Infers step size from replicates` wig files.

    Args:
        wig_path (str) : path to *directory* containing wig files


    Returns:
        int: an integer locus/step size used in the wig files
    """
    i = 0
    pos1 = 0
    pos2 = 0
    wig_file = [x for x in os.listdir(wig_path) if '.wig' in x][0]
    with open(wig_path + '/' + wig_file, mode='r', encoding='utf-8') as wig:
        for line in wig:
            line = line.strip()
            if line.replace('\t', '').replace(' ', '').isnumeric():
                if i == 0:
                    pos1 = int(line.split('\t')[0])
                if i == 1:
                    pos2 = int(line.split('\t')[0])
                if pos2 - pos1 > 0:
                    break
                i += 1
    return pos2 - pos1

def get_start_end(wig_path):
    """
    Infers starting and ending nucleotide position common to the
    multiple signal tracks

    To infer a reasonable genomic region over which to run ROCCO,
    this function returns the median starting point and endpoints
    among the multiple replicates' signal files in `wig_path`.

    Args:
        wig_path (str) : path to *directory* containing wig files

    Returns:
        start (int): median starting position
        end (int): median ending position
    """
    starts = []
    ends = []

    def get_start(track):
        with open(track, mode='r', encoding='utf-8') as wig:
            for line in wig:
                line = line.strip()

                if line.replace('\t', '').replace(' ', '').isnumeric():
                    start = int(line.split('\t')[0])
                    return start
        return None

    def get_end(track):
        with subprocess.Popen(["tail", "-n1", track],
                              stdout=subprocess.PIPE) as proc:
            for line in proc.stdout:
                line = line.decode()
                line = line.strip()
                return int(line.split('\t')[0])
        return None

    wig_files = [x for x in os.listdir(wig_path) if '.wig' in x]
    for wig_file in wig_files:
        starts.append(get_start(wig_path + '/' + wig_file))
        ends.append(get_end(wig_path + '/' + wig_file))
    return sorted(starts)[len(starts) // 2], sorted(ends)[len(ends) // 2]


def read_wig(wig_file, start=0, end=10**10, locus_size=50):
    r"""
    Processes a wig file for inclusion in the signal matrix $\mathbf{S}_{chr}$

    This function prepares signal data for to fit in a constant-sized
    $K \times n$ matrix. If a wig file begins at a position < `start`,
    it will be padded with zeros. Likewise if a wig file ends at a
    position > `end`, the extra values are ignored. Gaps between `start`
    and `end` are replaced with a zero-entry.

    Args:
        wig_file (str): path to wiggle formatted signal track
        start (int): inferred or manually-specified starting nucleotide
          position
        end (int): inferred or manually-specified ending nucleotide
          position
        locus_size (int): interval/locus size. Each wig file will have
          signal values at every `start + i*locus_size` nucleotide pos.
          for i = 0,1,2,...

    Returns:
        loci: list of starting nucleotide positions for each locus
        signal: enrichment signal value at each locus in `loci`
    """
    print("reading: {}".format(wig_file))
    loci = []
    signal = []
    with open(wig_file, mode='r', encoding='utf-8') as wig:
        for line in wig:
            line = line.strip()
            try:
                line = line.split('\t')
                line[0] = int(line[0])
                line[1] = float(line[1])
                if line[1] < 0:
                    line[1] = 0
                # case: wig begins after specified `start`
                if start < line[0] and len(loci) == 0:
                    for loc in range(start, line[0], locus_size):
                        loci.append(loc)
                        signal.append(0)
                    loci.append(line[0])
                    signal.append(line[1])
                    continue
                if line[0] < start:
                    continue
                if line[0] > end:
                    break
                # case: loci not contiguous
                if len(loci) > 0 and line[0] - loci[-1] > locus_size:
                    loci.append(loci[-1] + locus_size)
                    signal.append(0)
                    continue

                loci.append(line[0])
                signal.append(line[1])

            except ValueError:
                continue
            except IndexError:
                continue
    # case: wig ends prematurely
    if loci[-1] < end:
        for loc in range(loci[-1] + locus_size, end + locus_size, locus_size):
            loci.append(loc)
            signal.append(0)

    return loci, signal


def collect_wigs(wig_files, start=0, end=10**10, locus_size=50) -> pd.DataFrame:
    r"""
    creates a $K \times n$ dataframe of signals derived from wig files

    Args:
        wig_files (list): list of wig filepaths
        start (int): starting nucleotide position of wig files
        end (int): ending nucleotide position of wig files
        locus_size (int): initial size (in nucleotides) of Locus objects.
            Corresponds to $L$ in paper.

    Returns:
        pd.DataFrame: $K \times n$ dataframe of individuals' signals.
    """
    loci = [x for x in range(start, end + locus_size, locus_size)]
    all_wigs = pd.DataFrame(
        np.array([read_wig(x, start, end, locus_size)[1]
                  for x in wig_files],dtype=np.float32),
        columns=loci, index=None)
    return all_wigs


def emp_cdf(val, arr) -> float:
    """
    Get empirical cdf of `val` in `arr`

    Used to compute 0-1000 scaled score for BED6 files
    and visualization in genome browser.

    Note:
        Assumes `arr` is sorted!

    Args:
        val (float): a numeric element in `arr`
        arr (array): a SORTED array of peak values

    Returns:
        (float) estimate for percentile of `val` in `arr`
    """
    return np.searchsorted(arr, val) / len(arr)


def sig_selected(loci_obj: Loci) -> np.ndarray:
    """
    Get enrichment signals from all selected loci

    Args:
        loci_obj (Loci): Loci object after optimization

    Returns:
        np.ndarray: vector of enrichment signals from all selected
            (but still unmerged) loci
    """
    sig_ref = []
    for loc in loci_obj:
        if loc.accessible == 1:
            sig_ref.append(np.sum(loc.sig_data) / loc.size)
    return np.array(sig_ref)


def tags(fname) -> list:
    """
    Get sample IDs from file specified with command-line argument `--identifiers`

    Args:
        fname (str): path to file containing a subset of sample IDs--one line for each ID

    Returns:
        list: list of sample IDs in `fname`
    """
    tags_ = []
    for line in open(fname, mode='r', encoding='utf-8'):
        if line.strip().isspace() or len(line.strip()) == 0:
            continue
        tags_.append(line.strip())
    return tags_



def main(args):
    if args['verbose']:
        print(f'chrom: initial args\n{args}')
    if args['locus_size'] == -1:
        args['locus_size'] = get_locus_size(args['wig_path'])

    args['outdir'] = os.path.normpath(args['outdir'])
    if not os.path.exists(args['outdir']):
        os.mkdir(args['outdir'])

    if args['start'] == -1 or args['end'] == -1:
        args['start'], args['end'] = get_start_end(args['wig_path'])
    if args['bed_format'] not in [3, 6]:
        warnings.warn('Only BED3 and BED6 are supported--setting to BED6')
        args['bed_format'] = 6
    if args['verbose']:
        print(f'chrom: inferred args\n{args}')


    wig_files = [
        os.path.join(args['wig_path'], fname)
        for fname in os.listdir(args['wig_path'])
        if os.path.splitext(fname)[-1].lower() == '.wig'
    ]

    # select a subset of the wig files if `--identifiers` is defined
    if args['identifiers'] is not None:
        with open(os.path.join(args['outdir'], f"{args['chrom']}_sample_wig_match.log"), 'w') as samp_wig_log:
            new_wig_files = []
            for tag in tags(args['identifiers']):
                matching_wigs = []
                for wfname in wig_files:
                    if tag in wfname:
                        matching_wigs.append(wfname)
                tag_match = None
                if len(matching_wigs) > 1:
                    # pick the shortest match
                    tag_match = sorted(matching_wigs, key=len)[0]
                elif len(matching_wigs) == 1:
                    tag_match = matching_wigs[0]
                elif len(matching_wigs) == 0:
                    warnings.warn(f'chrom: no wig file found for sample identifier {tag}')
                if tag_match is not None:
                    print(f'chrom: sample/wig match: {tag}/{tag_match}')
                    samp_wig_log.write(f'{tag}\t{tag_match}\n')
                    new_wig_files.append(tag_match)
        wig_files = new_wig_files

    if len(wig_files) == 0:
        warnings.warn(f"chrom: did not any find wig files for the given samples + {args['chrom']}")
        sys.exit(0)

    # create a dataframe of signal values for `K` replicates and `n` loci
    signal_matrix = collect_wigs(wig_files,
                                 start=args['start'],
                                 end=args['end'],
                                 locus_size=args['locus_size'])

    InitLoci = Loci()
    for i, loc in enumerate(signal_matrix.columns):
        size_ = args['locus_size']
        NewLoc = Locus(position=args['start'] + i * size_,
                       size=size_,
                       sig_data=np.array(signal_matrix[loc],dtype=np.float32))
        InitLoci.append_locus(NewLoc)
    del signal_matrix
    gc.collect()

    InitLoci.rocco_lp(budget=args['budget'],
                        gam=args['gamma'],
                        tau=args['tau'],
                        c1=args['c1'],
                        c2=args['c2'],
                        c3=args['c3'],
                        verbose_=args['verbose'],
                        solver=args['solver'],
                        N=args['rr_iter'])

    InitLoci.combine_selected()

    fname = os.path.join(args['outdir'], "ROCCO_out_{}_{}_{}_{}_{}_{}_{}.bed".format(
        args['chrom'],
        args['budget'],
        args['gamma'],
        args['tau'],
        args['c1'],
        args['c2'],
        args['c3']))

    if os.path.exists(fname):
        warnings.warn(f'chrom: overwriting {fname}')
    print('chrom: writing output: {}'.format(fname))
    outfile = open(fname, mode='w', encoding='utf-8')
    output = ""
    if args['bed_format'] == 6:
        loc_count = 1
        sorted_sel_sig = np.sort(sig_selected(InitLoci))
        for loc in InitLoci:
            assert loc.size > 0
            if loc.accessible == 1:
                # the score can be used for coloring in ucsc genome browser
                loc_score = int(500 * emp_cdf(np.sum(loc.sig_data) / loc.size,
                                              sorted_sel_sig) + 500)
                loc_name = args['chrom'] + '_' + str(loc_count)
                output += "{}\t{}\t{}\t{}\t{}\t{}\n".format(
                    args['chrom'], loc.position, loc.position + loc.size,
                    loc_name, loc_score, '.')
                loc_count += 1

    if args['bed_format'] == 3:
        for loc in InitLoci:
            if loc.accessible == 1:
                output += "{}\t{}\t{}\n".format(
                    args['chrom'], loc.position, loc.position + loc.size)

    outfile.write(output)
    outfile.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--chrom', type=str, help="Specifies the chromosome, e.g., `--chrom chr1`.")
    parser.add_argument('--start', type=int, default=-1, help='beginning nucleotide position (default: infer from wiggle files)')
    parser.add_argument('--end', type=int, default=-1, help='ending nucleotide position (default: infer from wiggle files)')
    parser.add_argument('-L','--locus_size', type=int, default=-1, help='fixed step size in the samples wiggle tracks (default: infer from wiggle files)')
    parser.add_argument('--wig_path', type=str, default=os.getcwd(), help='Directory generated by `rocco prep` containing chromosome-specific .wig files for each sample, e.g., `--wig_path tracks_chr1`')
    parser.add_argument('-N', '--rr_iter', type=int, default=50, help = 'number of RR iterations')
    parser.add_argument('--verbose', default=False, action="store_true", help='set to `True` for verbose logging')
    parser.add_argument('-b', '--budget', type=float, default=.035, help='Upper-bounds the fraction of loci selected as open. Defines the budget constraint (Default: 0.035)')
    parser.add_argument('-g', '--gamma', type=float, default=1.0, help='fragmentation penalty in objective func. (Default: 1.0)')
    parser.add_argument('-t', '--tau', type=float, default=0.0, help='tau parameter (enrichment threshold)')
    parser.add_argument('--c1', type=float, default=1.0, help='c_1 coefficient in score function (enrichment reward)')
    parser.add_argument('--c2', type=float, default=1.0, help='c_2 coefficient in score function (dispersion penalty) ')
    parser.add_argument('--c3', type=float, default=1.0, help='c_3 coefficient in score function (local shift)')
    parser.add_argument('--solver', default="CLARABEL", help='Optimization software used to solve the \
                        main LP. `CLARABEL` is used by default.')
    parser.add_argument('--bed_format', type=int, default=6,
                        help="`3` for BED3 format and `6` for BED6 format")
    parser.add_argument('--identifiers', default=None,
                        help="text file containing a subset of sample IDs, one on each line, for samples to include in the experiment. If not invoked, defaults to `None`, and *all* samples with wig files in `--wig_path` are used.")
    parser.add_argument('--outdir', type=str, default='.', help="directory in which to store output bed file")

    args = vars(parser.parse_args())
    main(args)



"""
## [ROCCO](https://github.com/nolan-h-hamilton/ROCCO/): rocco_aux.py 
Miscellaneous helper functions
"""
import os
import subprocess
import pandas as pd
import pysam
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor

def sort_combine_bed(outfile: str, dir_: str = '.', exclude_list: list = ['EBV', 'M', 'MT']):
    """
    Sorts and combines chromosome-specific peak files. Creates a new
    bed file `outfile` to store the combined results.

    Parameters:
        outfile (str): output file name
        dir_ (str): directory containing chromosome-specific bed files
            Default is the current directory ('.').
        exclude_list (list): list of chromosomes to exclude. (default: ['EBV', 'M', 'MT'])

    Notes:
        - this function assumes the bed files are named with the\
            same convention and delimiter '_'
    """
    dir_ = os.path.normpath(dir_)

    filenames = [f_ for f_ in os.listdir(dir_)
                 if f_.split('.')[-1] == 'bed' and 'ROCCO_out' in f_
                 and f_ not in exclude_list]

    name_template = []
    try:
        name_template = filenames[0].split('_')
    except (IndexError, ValueError) as none_found:
        print('sort_combine_bed(): no valid ROCCO output files found in `dir_`')
        raise none_found
    chr_index = [i for i in range(len(name_template)) if 'chr' in name_template[i][:3]][0]
    filenames = sorted(filenames,
                       key=lambda x: int(val) if (val := x.split('_')[chr_index][3:]).isnumeric() else ord(val))
    filenames = [os.path.join(dir_, fname) for fname in filenames]
    with open(outfile, mode='w', encoding='utf-8') as outfile_:
        for fname in filenames:
            cat_process = subprocess.Popen(('cat', fname), stdout=outfile_.fileno())
            cat_process.wait()


def download_file(url, output_path):
    "download file at `url` to `output_path` using either curl or wget"
    try:
        subprocess.run(['curl', '-o', output_path, url], check=True)
    except subprocess.CalledProcessError:
        try:
            subprocess.run(['wget', '-O', output_path, url], check=True)
        except subprocess.CalledProcessError:
            raise RuntimeError("Failed to download the file.")


def get_size_file(assembly='hg38', exclude_list=['EBV', 'M', 'MT']) -> str:
    """
    if `assembly` is included in the UC Santa Cruz genome repository
    this function downloads, filters, and sorts the corresponding sizes
    file to `assembly`.sizes. Chromosomes in `exclude_list` are ignored.

    Args:
        assembly (str, optional): name of genome assembly. Defaults to 'hg38'.
        exclude_list (list, optional): list of chromosome names to exclude in sizes file.
            Defaults to ['EBV', 'M', 'MT'].

    Returns:
        str: path to sizes file, `assembly`.sizes
    """
    assembly = assembly.lower()
    url = f"http://hgdownload.soe.ucsc.edu/goldenPath/{assembly}/bigZips/{assembly}.chrom.sizes"
    output_path = f"{assembly}.sizes"
    print(f'\n{url} --> {output_path}\n')
    download_file(url, output_path)
    with open(output_path, 'r') as file_:
        lines = file_.readlines()
    filtered_lines = [line for line in lines if line.strip() and not any(excluded_chr in line for excluded_chr in exclude_list) and '_' not in line]
    sorted_lines = sorted(filtered_lines, key=lambda x: int(x.split('\t')[0][3:]) if x.split('\t')[0][3:].isnumeric() else ord(x.split('\t')[0][3:]))
    sorted_output_path = f"{assembly}.sizes"
    with open(sorted_output_path, 'w') as file_:
        file_.writelines(sorted_lines)
    return sorted_output_path


def parse_size_file(size_file, exclude_list=['EBV', 'M', 'MT']) -> dict:
    """Parse a size file and return a dictionary {chr: size}.

    Args:
        size_file (str): Path to the size file.
        exclude_list (list): list of chromosomes to exclude. (default: ['EBV', 'M', 'MT'])
    Returns:
        dict: a dictionary where chromosome names are keys and their sizes are values.
    """
    try:
        df = pd.read_csv(size_file, sep='\t', header=None)
    except Exception as ex:
        print(f'\nrocco_aux.parse_size_file(): Could not parse {size_file}. Ensure it is a valid sizes file.\n')
        raise
    size_dict = {key: val
                 for key, val in OrderedDict(zip(df.iloc[:, 0], df.iloc[:, 1])).items()
                    if key not in exclude_list}
    return size_dict


def is_alignment(filepath) -> bool:
    """
    Check if a file is an alignment file.

    Args:
        filepath (str): The path to the file.

    Returns:
        bool: True if the file is an alignment file, False otherwise.
    """
    if filepath.split('.')[-1] in ['bam','sam']:
        try:
            pysam.head("-n 1", filepath)
            return True
        except pysam.SamtoolsError:
            return False
    return False

def run_par(cmd_file, threads=-1, verbose=False):
    """
    Runs shell commands in `cmd_file` in parallel

    Args:
        cmd_file (str): file containing newline-separated commands
        threads (int): number of threads to use. if default '-1',
            then the number of threads is set to 1 + os.cpu_count()//2
        verbose (bool): whether to print subprocess output to stdout
    """
    assert os.path.exists(cmd_file), f'supplied `cmd_file` could not be found'
    if threads == -1:
        threads = 1 + os.cpu_count()//2
    with open(cmd_file, 'r') as file:
        commands = file.readlines()

    commands = [cmd.strip() for cmd in commands][::-1]


    def run_command(command, verbose=verbose):
        if not verbose:
            process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            process.communicate()
            return command, process.returncode
        if verbose:
            process = subprocess.Popen(command, shell=True, stderr=subprocess.DEVNULL)
            process.communicate()
            return command, process.returncode

    with ThreadPoolExecutor(max_workers = threads) as executor:
        futures = [executor.submit(run_command, cmd) for cmd in commands]
        results = [future.result() for future in futures]
    if verbose:
        for command, returncode in results[::-1]:
            print(f"cmd: {command}\nretval: {returncode}\n")

def has_reads(bamfile, min_reads=1, chrom='chrY') -> bool:
    """
    Check if a BAM file contains a minimum number of reads for a specific chromosome.

    Args:
        bamfile (str): path to the BAM file
        min_reads (int): minimum read threshold
        chrom (str, optional): chromosome to check for reads

    Returns:
        bool: True if the BAM file contains at least 'min_reads' for 'chrom', False otherwise.
    """
    bamfile = pysam.AlignmentFile(bamfile, "rb")
    has_min_reads = False

    for elem in bamfile.get_index_statistics():
        if elem[0] == chrom and elem[1] >= min_reads:
            has_min_reads = True
            break
    bamfile.close()
    return has_min_reads



"""
## [ROCCO](https://github.com/nolan-h-hamilton/ROCCO/): budgets.py
Compute chromosome-specific budgets from sample data based on relative read densities.


Parameters:
    --wigdir (str): Specifies the path to the directory containing the tracks_chr[] subdirectories generated by `rocco prep`.
    --smean (float): The chromosome-specific budgets will scale to this value. Default is 0.035.
    --samp_rate (float): Fraction of loci to sample for computing the mean signal value.
    -s, --sizes (str): Chromosome sizes file. Default is 'hg38.sizes'.
    -o, --outfile (str): Name of the output file. Output CSV is structured for use with `rocco gwide --param_file`. Default is 'params.csv'.
    -L, --locsize (int): Step-size in wig files. Default is 50.

Input:
    Path to directory containing the `tracks_chr[]` output from `rocco prep`

Output:
    CSV file that can be supplied as the parameter file for `rocco gwide --param_file`

Example [from demo.ipynb](https://github.com/nolan-h-hamilton/ROCCO/blob/main/demo/demo.ipynb):
    ```
    rocco budgets --wigdir . -s hg38.sizes --smean .035 --samp_rate .20 -o params.csv
    ```

"""

import os
import argparse
import random
import numpy as np
from typing import Dict
import pysam
from . import rocco_aux
from . import chrom

def sample_wigs(chrom_wigdir, samp_rate=1.0, locsize=50) -> float:
    """
    Sample loci in `chrom`, compute mean signal values at each, compute mean over all sampled loci

    Args:
        chrom_wigdir (str): The path to the chromosome-specific wig directory. e.g., `tracks_chr1`
        samp_rate (float): Fraction of loci to sample for computing the mean signal value.
        locsize (int): Step-size in wig files.

    Returns:
        float: mean signal value normalized by the sampling rate and the number of loci.

    """
    chrom_wigdir = rocco_aux.trim_path(chrom_wigdir)
    wig_files = [os.path.join(chrom_wigdir, fname)
                 for fname in os.listdir(chrom_wigdir) if 'wig' in fname.split('.')[-1]]
    start_, end_ = chrom.get_start_end(chrom_wigdir)

    signal_matrix = chrom.collect_wigs(wig_files,
                                 start=start_,
                                 end=end_,
                                 locus_size=locsize)
    sum_=0
    for i, loc in enumerate(signal_matrix.columns):
        if random.random() <= samp_rate:
            sigs = np.array(signal_matrix[loc],dtype=np.float32)
            sum_ += np.mean(sigs)
    return sum_/(samp_rate*len(signal_matrix.columns))

def scale_densities(chrom_dict, smean=.035) -> dict:
    """
    Scale chromosome-specific densities to mean `smean`.

    Args:
        chrom_dict (Dict): A dictionary mapping chromosome names to density values.
        smean (float): The target mean value to which densities will be scaled.

    Returns:
        Dict: A dictionary with chromosome names as keys and budget estimates averaging to `smean`

    """
    curr_mean = np.mean([float(x) for x in chrom_dict.values()])
    for key in chrom_dict.keys():
        chrom_dict[key] *= smean/curr_mean
    return chrom_dict

def main(args):

    chrom_dict = rocco_aux.parse_size_file(args['sizes'])
    args['wigdir'] = rocco_aux.trim_path(args['wigdir'])
    for key in chrom_dict.keys():
        chrom_dict[key] = sample_wigs(args['wigdir'] + "/tracks_" + str(key), samp_rate=args['samp_rate'],locsize=args['locsize'])
    chrom_dict = scale_densities(chrom_dict,smean=args['smean'])
    out = open(args['outfile'],'w')
    out.write('chromosome,input_path,budget,gamma,tau,c1,c2,c3\n')
    for key, val in chrom_dict.items():
        out.write(f'{key},{args["wigdir"] + "/tracks_" + str(key)},{round(val, 3)},NULL,NULL,NULL,NULL,NULL\n')
    out.close()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Compute chromosome-specific budget parameters based on read densities')
    parser.add_argument('--wigdir',type=str, help="specifies the path to the directory containing the tracks_chr[] subdirectories generated by `rocco prep`" )
    parser.add_argument('--smean', type=float, default=.035, help="the chromosome-specific budgets will scale to this value")
    parser.add_argument('--samp_rate', type=float, help="fraction of loci to sample for computing mean signal value")
    parser.add_argument('-s','--sizes', type=str, default='hg38.sizes', help='chromosome sizes file')
    parser.add_argument('-o', '--outfile', type=str, default='params.csv', help='name of output file. Output CSV is structured for use with `rocco gwide --param_file`')
    parser.add_argument('-L','--locsize', type=int, default=50, help='step-size in wig files')
    args = vars(parser.parse_args())
    main(args)
